@@ -1,21 +1,77 @@
-#include "ecore_private.h"
-#include "Ecore.h"
+/*
+ * vim:ts=8:sw=3:sts=8:noexpandtab:cino=>5n-3f0^-2{2
+ */
 
-static int                  events_num = 0;
-static Ecore_Event         *events = NULL;
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+
+#include <stdlib.h>
+
+#include "Ecore.h"
+#include "ecore_private.h"
+
+static int inpurge = 0;
+
+struct _Ecore_Event_Handler
+{
+   EINA_INLIST;
+   ECORE_MAGIC;
+   int type;
+   Eina_Bool (*func) (void *data, int type, void *event);
+   void *data;
+   int       references;
+   Eina_Bool delete_me : 1;
+};
+
+struct _Ecore_Event_Filter
+{
+   EINA_INLIST;
+   ECORE_MAGIC;
+   void *(*func_start) (void *data);
+   Eina_Bool (*func_filter) (void *data, void *loop_data, int type, void *event);
+   void (*func_end) (void *data, void *loop_data);
+   void *loop_data;
+   void *data;
+   int       references;
+   Eina_Bool delete_me : 1;
+};
+
+struct _Ecore_Event
+{
+   EINA_INLIST;
+   ECORE_MAGIC;
+   int type;
+   void *event;
+   void (*func_free) (void *data, void *ev);
+   void *data;
+   int       references;
+   Eina_Bool delete_me : 1;
+};
+
+
+static int events_num = 0;
+static Ecore_Event *events = NULL;
+static Ecore_Event *event_current = NULL;
+static Ecore_Event *purge_events = NULL;
 
 static Ecore_Event_Handler **event_handlers = NULL;
-static int                   event_handlers_num = 0;
-static int                   event_handlers_alloc_num = 0;
-static Ecore_List2_Data     *event_handlers_delete_list = NULL;
+static Ecore_Event_Handler *event_handler_current = NULL;
+static int event_handlers_num = 0;
+static int event_handlers_alloc_num = 0;
+static Eina_List *event_handlers_delete_list = NULL;
 
-static Ecore_Event_Filter  *event_filters = NULL;
-static int                  event_filters_delete_me = 0;
+static Ecore_Event_Filter *event_filters = NULL;
+static Ecore_Event_Filter *event_filter_current = NULL;
+static Ecore_Event *event_filter_event_current = NULL;
+static int event_filters_delete_me = 0;
+static int event_id_max = ECORE_EVENT_COUNT;
+static int ecore_raw_event_type = ECORE_EVENT_NONE;
+static void *ecore_raw_event_event =  NULL;
 
-static int                  event_id_max = ECORE_EVENT_COUNT;
 
-static int                  ecore_raw_event_type = ECORE_EVENT_NONE;
-static void                *ecore_raw_event_event =  NULL;
+static void _ecore_event_purge_deleted(void);
+static void *_ecore_event_del(Ecore_Event *event);
 
 
 /**
@@ -24,25 +80,25 @@ static void                *ecore_raw_event_event =  NULL;
  * @param func The function to call when the event is found in the queue
  * @param data A data pointer to pass to the called function @p func
  * @return A new Event handler, or NULL on failure
- * 
+ *
  * Add an event handler to the list of handlers. This will, on success, return
  * a handle to the event handler object that was created, that can be used
- * later to remove the handler using ecore_event_handler_del(). The @p type 
- * parameter is the iteger of the event type that will trigger this callback
+ * later to remove the handler using ecore_event_handler_del(). The @p type
+ * parameter is the integer of the event type that will trigger this callback
  * to be called. The callback @p func is called when this event is processed
  * and will be passed the event type, a pointer to the private event
- * structure that is specific to that event type, and a data pointer that is 
+ * structure that is specific to that event type, and a data pointer that is
  * provided in this call as the @p data parameter.
- * 
- * When the callback @p func is called, it must return 1 or 0. If it returns 
- * 1 (or ECORE_CALLBACK_RENEW), It will keep being called as per normal, for 
- * each handler set up for that event type. If it returns 0 (or 
+ *
+ * When the callback @p func is called, it must return 1 or 0. If it returns
+ * 1 (or ECORE_CALLBACK_RENEW), It will keep being called as per normal, for
+ * each handler set up for that event type. If it returns 0 (or
  * ECORE_CALLBACK_CANCEL), it will cease processing handlers for that particular
- * event, so all handler set to handle that event type that have not already 
+ * event, so all handler set to handle that event type that have not already
  * been called, will not be.
  */
 EAPI Ecore_Event_Handler *
-ecore_event_handler_add(int type, int (*func) (void *data, int type, void *event), const void *data)
+ecore_event_handler_add(int type, Eina_Bool (*func) (void *data, int type, void *event), const void *data)
 {
    Ecore_Event_Handler *eh;
 
@@ -57,14 +113,14 @@ ecore_event_handler_add(int type, int (*func) (void *data, int type, void *event
    if (type >= (event_handlers_num - 1))
      {
 	int p_alloc_num;
-	
+
 	p_alloc_num = event_handlers_alloc_num;
 	event_handlers_num = type + 1;
 	if (event_handlers_num > event_handlers_alloc_num)
 	  {
 	     Ecore_Event_Handler **new_handlers;
 	     int i;
-	     
+
 	     event_handlers_alloc_num = ((event_handlers_num + 16) / 16) * 16;
 	     new_handlers = realloc(event_handlers, event_handlers_alloc_num * sizeof(Ecore_Event_Handler *));
 	     if (!new_handlers)
@@ -77,7 +133,7 @@ ecore_event_handler_add(int type, int (*func) (void *data, int type, void *event
 	       event_handlers[i] = NULL;
 	  }
      }
-   event_handlers[type] = _ecore_list2_append(event_handlers[type], eh);
+   event_handlers[type] = (Ecore_Event_Handler *) eina_inlist_append(EINA_INLIST_GET(event_handlers[type]), EINA_INLIST_GET(eh));
    return eh;
 }
 
@@ -85,7 +141,7 @@ ecore_event_handler_add(int type, int (*func) (void *data, int type, void *event
  * Delete an event handler.
  * @param event_handler Event handler handle to delete
  * @return Data passed to handler
- * 
+ *
  * Delete a specified event handler from the handler list. On success this will
  * delete the event handler and return the pointer passed as @p data when the
  * handler was added by ecore_event_handler_add(). On failure NULL will be
@@ -94,22 +150,18 @@ ecore_event_handler_add(int type, int (*func) (void *data, int type, void *event
 EAPI void *
 ecore_event_handler_del(Ecore_Event_Handler *event_handler)
 {
-   Ecore_List2_Data *node;
-   
-   if (!ECORE_MAGIC_CHECK(event_handler, ECORE_MAGIC_EVENT_HANDLER)) 
+   if (!ECORE_MAGIC_CHECK(event_handler, ECORE_MAGIC_EVENT_HANDLER))
      {
 	ECORE_MAGIC_FAIL(event_handler, ECORE_MAGIC_EVENT_HANDLER,
 			 "ecore_event_handler_del");
 	return NULL;
      }
    event_handler->delete_me = 1;
-   node = calloc(1, sizeof(Ecore_List2_Data));
-   node->data = event_handler;
-   event_handlers_delete_list = _ecore_list2_append(event_handlers_delete_list, node);
+   event_handlers_delete_list = eina_list_append(event_handlers_delete_list, event_handler);
    return event_handler->data;
 }
 
-static void 
+static void
 _ecore_event_generic_free (void *data __UNUSED__, void *event)
 {
    free (event);
@@ -122,10 +174,10 @@ _ecore_event_generic_free (void *data __UNUSED__, void *event)
  * @param func_free The function to be called to free this private structure
  * @param data The data pointer to be passed to the free function
  * @return A Handle for that event
- * 
+ *
  * On success this function returns a handle to an event on the event queue, or
  * NULL if it fails. If it succeeds, an event of type @p type will be added
- * to the queue for processing by event handlers added by 
+ * to the queue for processing by event handlers added by
  * ecore_event_handler_add(). The @p ev parameter will be a pointer to the event
  * private data that is specific to that event type. When the event is no
  * longer needed, @p func_free will be called and passed the private structure
@@ -147,7 +199,7 @@ ecore_event_add(int type, void *ev, void (*func_free) (void *data, void *ev), vo
  * Delete an event from the queue.
  * @param event The event handle to delete
  * @return The data pointer originally set for the event free function
- * 
+ *
  * This deletes the event @p event from the event queue, and returns the
  * @p data parameer originally set when adding it with ecore_event_add(). This
  * does not immediately call the free function, and it may be called later on
@@ -157,9 +209,9 @@ ecore_event_add(int type, void *ev, void (*func_free) (void *data, void *ev), vo
 EAPI void *
 ecore_event_del(Ecore_Event *event)
 {
-   if (!ECORE_MAGIC_CHECK(event, ECORE_MAGIC_EVENT)) 
+   if (!ECORE_MAGIC_CHECK(event, ECORE_MAGIC_EVENT))
      {
-	ECORE_MAGIC_FAIL(event, ECORE_MAGIC_EVENT, "ecore_event_del");	
+	ECORE_MAGIC_FAIL(event, ECORE_MAGIC_EVENT, "ecore_event_del");
 	return NULL;
      }
    event->delete_me = 1;
@@ -169,8 +221,8 @@ ecore_event_del(Ecore_Event *event)
 /**
  * Allocate a new event type id sensibly and return the new id.
  * @return A new event type id.
- * 
- * This function allocates a new event type id and returns it. Once an event 
+ *
+ * This function allocates a new event type id and returns it. Once an event
  * type has been allocated it can never be de-allocated during the life of
  * the program. There is no guarantee of the contents of this event ID, or how
  * it is calculated, except that the ID will be unique to the current instance
@@ -190,8 +242,8 @@ ecore_event_type_new(void)
  * @param func_end Function to call after the queu has been filtered
  * @param data Data to pass to the filter functions
  * @return A filter handle
- * 
- * This adds a filter to call callbacks to loop through the event queue and 
+ *
+ * This adds a filter to call callbacks to loop through the event queue and
  * filter events out of the queue. On failure NULL is returned. On success a
  * Filter handle is returned. Filters are called on the queue just before
  * Event handler processing to try and remove redundant events. Just as
@@ -204,10 +256,10 @@ ecore_event_type_new(void)
  * and @p data pointer to clean up.
  */
 EAPI Ecore_Event_Filter *
-ecore_event_filter_add(void * (*func_start) (void *data), int (*func_filter) (void *data, void *loop_data, int type, void *event), void (*func_end) (void *data, void *loop_data), const void *data)
+ecore_event_filter_add(void * (*func_start) (void *data), Eina_Bool (*func_filter) (void *data, void *loop_data, int type, void *event), void (*func_end) (void *data, void *loop_data), const void *data)
 {
    Ecore_Event_Filter *ef;
-   
+
    if (!func_filter) return NULL;
    ef = calloc(1, sizeof(Ecore_Event_Filter));
    if (!ef) return NULL;
@@ -216,7 +268,7 @@ ecore_event_filter_add(void * (*func_start) (void *data), int (*func_filter) (vo
    ef->func_filter = func_filter;
    ef->func_end = func_end;
    ef->data = (void *)data;
-   event_filters = _ecore_list2_append(event_filters, ef);
+   event_filters = (Ecore_Event_Filter *) eina_inlist_append(EINA_INLIST_GET(event_filters), EINA_INLIST_GET(ef));
    return ef;
 }
 
@@ -224,14 +276,14 @@ ecore_event_filter_add(void * (*func_start) (void *data), int (*func_filter) (vo
  * Delete an event filter.
  * @param ef The event filter handle
  * @return The data set for the filter
- * 
+ *
  * Delete a filter that has been added by its @p ef handle. On success this
  * will return the data pointer set when this filter was added. On failure
  * NULL is returned.
  */
 EAPI void *
 ecore_event_filter_del(Ecore_Event_Filter *ef)
-{   
+{
    if (!ECORE_MAGIC_CHECK(ef, ECORE_MAGIC_EVENT_FILTER))
      {
 	ECORE_MAGIC_FAIL(ef, ECORE_MAGIC_EVENT_FILTER, "ecore_event_filter_del");
@@ -245,11 +297,11 @@ ecore_event_filter_del(Ecore_Event_Filter *ef)
 /**
  * Return the current event type being handled.
  * @return The current event type being handled if inside a handler callback
- * 
+ *
  * If the program is currently inside an Ecore event handler callback this
  * will return the type of the current event being processed. If Ecore is
  * not inside an event handler, ECORE_EVENT_NONE is returned.
- * 
+ *
  * This is useful when certain Ecore modules such as Ecore_Evas "swallow"
  * events and not all the original information is passed on. In special cases
  * this extra information may be useful or needed and using this call can let
@@ -265,11 +317,11 @@ ecore_event_current_type_get(void)
 /**
  * Return the current event type pointer handled.
  * @return The current event pointer being handled if inside a handler callback
- * 
+ *
  * If the program is currently inside an Ecore event handler callback this
  * will return the pointer of the current event being processed. If Ecore is
  * not inside an event handler, NULL will be returned.
- * 
+ *
  * This is useful when certain Ecore modules such as Ecore_Evas "swallow"
  * events and not all the original information is passed on. In special cases
  * this extra information may be useful or needed and using this call can let
@@ -286,48 +338,43 @@ void
 _ecore_event_shutdown(void)
 {
    int i;
-   
+   Ecore_Event_Handler *eh;
+   Ecore_Event_Filter *ef;
+
    while (events) _ecore_event_del(events);
+   event_current = NULL;
    for (i = 0; i < event_handlers_num; i++)
      {
-	while (event_handlers[i])
+	while ((eh = event_handlers[i]))
 	  {
-	     Ecore_Event_Handler *eh;
-	     
-	     eh = event_handlers[i];
-	     event_handlers[i] = _ecore_list2_remove(event_handlers[i], eh);
+	     event_handlers[i] = (Ecore_Event_Handler *) eina_inlist_remove(EINA_INLIST_GET(event_handlers[i]), EINA_INLIST_GET(event_handlers[i]));
 	     ECORE_MAGIC_SET(eh, ECORE_MAGIC_NONE);
-	     free(eh);
+	     if (!eh->delete_me) free(eh);
 	  }
      }
-   while (event_handlers_delete_list)
-     {
-	Ecore_List2_Data *ehd;
-	
-	ehd = event_handlers_delete_list;
-	event_handlers_delete_list = _ecore_list2_remove(event_handlers_delete_list, ehd);
-	free(ehd);
-     }
+   EINA_LIST_FREE(event_handlers_delete_list, eh)
+     free(eh);
    if (event_handlers) free(event_handlers);
    event_handlers = NULL;
    event_handlers_num = 0;
    event_handlers_alloc_num = 0;
-   while (event_filters)
+   while ((ef = event_filters))
      {
-	Ecore_Event_Filter *ef;
-	
-	ef = event_filters;
-	event_filters = _ecore_list2_remove(event_filters, ef);   
+	event_filters = (Ecore_Event_Filter *) eina_inlist_remove(EINA_INLIST_GET(event_filters), EINA_INLIST_GET(event_filters));
 	ECORE_MAGIC_SET(ef, ECORE_MAGIC_NONE);
 	free(ef);
      }
    event_filters_delete_me = 0;
+   event_filter_current = NULL;
+   event_filter_event_current = NULL;
 }
 
 int
 _ecore_event_exist(void)
 {
-   if (events) return 1;
+   Ecore_Event *e;
+   EINA_INLIST_FOREACH(events, e)
+      if (!e->delete_me) return 1;
    return 0;
 }
 
@@ -335,7 +382,7 @@ Ecore_Event *
 _ecore_event_add(int type, void *ev, void (*func_free) (void *data, void *ev), void *data)
 {
    Ecore_Event *e;
-   
+
    e = calloc(1, sizeof(Ecore_Event));
    if (!e) return NULL;
    ECORE_MAGIC_SET(e, ECORE_MAGIC_EVENT);
@@ -343,8 +390,16 @@ _ecore_event_add(int type, void *ev, void (*func_free) (void *data, void *ev), v
    e->event = ev;
    e->func_free = func_free;
    e->data = data;
-   events = _ecore_list2_append(events, e);
-   events_num++;
+   if (inpurge > 0)
+     {
+        purge_events = (Ecore_Event *)eina_inlist_append(EINA_INLIST_GET(purge_events), EINA_INLIST_GET(e));
+        events_num++;
+     }
+   else
+     {
+        events = (Ecore_Event *)eina_inlist_append(EINA_INLIST_GET(events), EINA_INLIST_GET(e));
+        events_num++;
+     }
    return e;
 }
 
@@ -352,104 +407,213 @@ void *
 _ecore_event_del(Ecore_Event *event)
 {
    void *data;
-   
+
    data = event->data;
    if (event->func_free) event->func_free(event->data, event->event);
-   events = _ecore_list2_remove(events, event);
+   events = (Ecore_Event *) eina_inlist_remove(EINA_INLIST_GET(events), EINA_INLIST_GET(event));
    ECORE_MAGIC_SET(event, ECORE_MAGIC_NONE);
    free(event);
    events_num--;
    return data;
 }
 
-void
-_ecore_event_call(void)
+static void
+_ecore_event_purge_deleted(void)
 {
-   Ecore_List2 *l, *ll;
-   Ecore_Event *e;
-   Ecore_Event_Filter *ef;
-   Ecore_Event_Handler *eh;
-   Ecore_List2_Data *ehd;
-   int handle_count;
+   Ecore_Event *itr = events;
 
-   for (l = (Ecore_List2 *)event_filters; l; l = l->next)
+   inpurge++;
+   while (itr)
      {
-	ef = (Ecore_Event_Filter *)l;
+	Ecore_Event *next = (Ecore_Event *)EINA_INLIST_GET(itr)->next;
+	if ((!itr->references) && (itr->delete_me))
+	  _ecore_event_del(itr);
+	itr = next;
+     }
+   inpurge--;
+   while (purge_events)
+     {
+        Ecore_Event *e = purge_events;
+        purge_events = (Ecore_Event *)eina_inlist_remove(EINA_INLIST_GET(purge_events), EINA_INLIST_GET(purge_events));
+        events = (Ecore_Event *)eina_inlist_append(EINA_INLIST_GET(events), EINA_INLIST_GET(e));
+     }
+}
+
+static inline void
+_ecore_event_filters_apply()
+{
+
+   if (!event_filter_current)
+     {
+	/* regular main loop, start from head */
+	event_filter_current = event_filters;
+     }
+   else
+     {
+	/* recursive main loop, continue from where we were */
+	event_filter_current = (Ecore_Event_Filter *)EINA_INLIST_GET(event_filter_current)->next;
+     }
+
+   while (event_filter_current)
+     {
+	Ecore_Event_Filter *ef = event_filter_current;
+
 	if (!ef->delete_me)
 	  {
+	     ef->references++;
+
 	     if (ef->func_start)
 	       ef->loop_data = ef->func_start(ef->data);
-	     for (ll = (Ecore_List2 *)events; ll; ll = ll->next)
+
+	     if (!event_filter_event_current)
 	       {
-		  e = (Ecore_Event *)ll;
-		  if (!ef->func_filter(ef->loop_data, ef->data,
+		  /* regular main loop, start from head */
+		  event_filter_event_current = events;
+	       }
+	     else
+	       {
+		  /* recursive main loop, continue from where we were */
+		  event_filter_event_current = (Ecore_Event *)EINA_INLIST_GET(event_filter_event_current)->next;
+	       }
+
+	     while (event_filter_event_current)
+	       {
+		  Ecore_Event *e = event_filter_event_current;
+
+		  if (!ef->func_filter(ef->data, ef->loop_data,
 				       e->type, e->event))
 		    {
-//		       printf("FILTER SAID TO DEL ev %p\n", e->event);
 		       ecore_event_del(e);
 		    }
+
+		  if (event_filter_event_current) /* may have changed in recursive main loops */
+		    event_filter_event_current = (Ecore_Event *)EINA_INLIST_GET(event_filter_event_current)->next;
 	       }
 	     if (ef->func_end)
 	       ef->func_end(ef->data, ef->loop_data);
+
+	     ef->references--;
 	  }
+
+	if (event_filter_current) /* may have changed in recursive main loops */
+	  event_filter_current = (Ecore_Event_Filter *)EINA_INLIST_GET(event_filter_current)->next;
      }
    if (event_filters_delete_me)
      {
-	for (l = (Ecore_List2 *)event_filters; l;)
+	int deleted_in_use = 0;
+	Ecore_Event_Filter *l;
+	for (l = event_filters; l;)
 	  {
-	     ef = (Ecore_Event_Filter *)l;
-	     l = l->next;
+	     Ecore_Event_Filter *ef = l;
+	     l = (Ecore_Event_Filter *) EINA_INLIST_GET(l)->next;
 	     if (ef->delete_me)
 	       {
-		  event_filters = _ecore_list2_remove(event_filters, ef);
+		  if (ef->references)
+		    {
+		       deleted_in_use++;
+		       continue;
+		    }
+
+		  event_filters = (Ecore_Event_Filter *) eina_inlist_remove(EINA_INLIST_GET(event_filters), EINA_INLIST_GET(ef));
 		  ECORE_MAGIC_SET(ef, ECORE_MAGIC_NONE);
 		  free(ef);
 	       }
 	  }
-	event_filters_delete_me = 0;
+	if (!deleted_in_use)
+	  event_filters_delete_me = 0;
      }
-//   printf("EVENT BATCH...\n");
-   for (l = (Ecore_List2 *)events; l; l = l->next)
+}
+void
+_ecore_event_call(void)
+{
+   Eina_List *l, *l_next;
+   Ecore_Event_Handler *eh;
+
+   _ecore_event_filters_apply();
+
+   if (!event_current)
      {
-	e = (Ecore_Event *)l;
-	if (!e->delete_me)
+	/* regular main loop, start from head */
+	event_current = events;
+	event_handler_current = NULL;
+     }
+
+   while (event_current)
+     {
+	Ecore_Event *e = event_current;
+	int handle_count = 0;
+
+	if (e->delete_me)
 	  {
-	     handle_count = 0;
-	     ecore_raw_event_type = e->type;
-	     ecore_raw_event_event = e->event;
-//	     printf("HANDLE ev type %i, %p\n", e->type, e->event);
-	     if ((e->type >= 0) && (e->type < event_handlers_num))
+	     event_current = (Ecore_Event *)EINA_INLIST_GET(event_current)->next;
+	     continue;
+	  }
+
+	ecore_raw_event_type = e->type;
+	ecore_raw_event_event = e->event;
+	e->references++;
+	if ((e->type >= 0) && (e->type < event_handlers_num))
+	  {
+	     if (!event_handler_current)
 	       {
-		  for (ll = (Ecore_List2 *)event_handlers[e->type]; ll; ll = ll->next)
+		  /* regular main loop, start from head */
+		  event_handler_current = event_handlers[e->type];
+	       }
+	     else
+	       {
+		  /* recursive main loop, continue from where we were */
+		  event_handler_current = (Ecore_Event_Handler *)EINA_INLIST_GET(event_handler_current)->next;
+	       }
+
+	     while ((event_handler_current) && (!e->delete_me))
+	       {
+		  Ecore_Event_Handler *eh = event_handler_current;
+		  if (!eh->delete_me)
 		    {
-		       eh = (Ecore_Event_Handler *)ll;
-		       if (!eh->delete_me)
+		       Eina_Bool ret;
+
+		       handle_count++;
+
+		       eh->references++;
+		       ret = eh->func(eh->data, e->type, e->event);
+		       eh->references--;
+
+		       if (!ret)
 			 {
-			    handle_count++;
-			    if (!eh->func(eh->data, e->type, e->event))
-			      break;  /* 0 == "call no further handlers" */
+			    event_handler_current = NULL;
+			    break;  /* 0 == "call no further handlers" */
 			 }
 		    }
+
+		  if (event_handler_current) /* may have changed in recursive main loops */
+		    event_handler_current = (Ecore_Event_Handler *)EINA_INLIST_GET(event_handler_current)->next;
 	       }
-	     /* if no handlers were set for EXIT signal - then default is */
-	     /* to quit the main loop */
-	     if ((e->type == ECORE_EVENT_SIGNAL_EXIT) && (handle_count == 0))
-	       ecore_main_loop_quit();
 	  }
+	/* if no handlers were set for EXIT signal - then default is */
+	/* to quit the main loop */
+	if ((e->type == ECORE_EVENT_SIGNAL_EXIT) && (handle_count == 0))
+	  ecore_main_loop_quit();
+	e->references--;
+	e->delete_me = 1;
+
+	if (event_current) /* may have changed in recursive main loops */
+	  event_current = (Ecore_Event *)EINA_INLIST_GET(event_current)->next;
      }
+
    ecore_raw_event_type = ECORE_EVENT_NONE;
    ecore_raw_event_event = NULL;
-   
-   while (events) _ecore_event_del(events);
-   while (event_handlers_delete_list)
+
+   _ecore_event_purge_deleted();
+
+   EINA_LIST_FOREACH_SAFE(event_handlers_delete_list, l, l_next, eh)
      {
-	ehd = event_handlers_delete_list;
-	eh = ehd->data;
-	event_handlers[eh->type] = _ecore_list2_remove(event_handlers[eh->type], eh);
-	event_handlers_delete_list = _ecore_list2_remove(event_handlers_delete_list, ehd);
+	if (eh->references) continue;
+
+	event_handlers_delete_list = eina_list_remove_list(event_handlers_delete_list, l);
+
+	event_handlers[eh->type] = (Ecore_Event_Handler *) eina_inlist_remove(EINA_INLIST_GET(event_handlers[eh->type]), EINA_INLIST_GET(eh));
 	ECORE_MAGIC_SET(eh, ECORE_MAGIC_NONE);
 	free(eh);
-	free(ehd);
      }
 }
 
@@ -457,7 +621,7 @@ EAPI void *
 _ecore_event_signal_user_new(void)
 {
    Ecore_Event_Signal_User *e;
-   
+
    e = calloc(1, sizeof(Ecore_Event_Signal_User));
    return e;
 }
@@ -466,7 +630,7 @@ void *
 _ecore_event_signal_hup_new(void)
 {
    Ecore_Event_Signal_Hup *e;
-   
+
    e = calloc(1, sizeof(Ecore_Event_Signal_Hup));
    return e;
 }
@@ -475,7 +639,7 @@ void *
 _ecore_event_signal_exit_new(void)
 {
    Ecore_Event_Signal_Exit *e;
-   
+
    e = calloc(1, sizeof(Ecore_Event_Signal_Exit));
    return e;
 }
@@ -484,7 +648,7 @@ void *
 _ecore_event_signal_power_new(void)
 {
    Ecore_Event_Signal_Power *e;
-   
+
    e = calloc(1, sizeof(Ecore_Event_Signal_Power));
    return e;
 }
