@@ -7,8 +7,8 @@
 #endif
 
 #if defined (__FreeBSD__) || defined (__OpenBSD__) || defined (__NetBSD__)
-#include <sys/time.h>
-#include <sys/resource.h>
+# include <sys/time.h>
+# include <sys/resource.h>
 #endif
 
 #include <stdlib.h>
@@ -23,10 +23,107 @@
 # include <sys/wait.h>
 #endif
 
-#include "ecore_private.h"
 #include "Ecore.h"
+#include "ecore_private.h"
+
+
+   /* FIXME: Getting respawn to work
+    *
+    * There is no way that we can do anything about the internal state info of
+    * an external exe.  The same can be said about the state of user code.  User
+    * code in this context means the code that is using ecore_exe to manage exe's
+    * for it.
+    *
+    * Document that the exe must be respawnable, in other words, there is no
+    * state that it cannot regenerate by just killing it and starting it again.
+    * This includes state that the user code knows about, as the respawn is
+    * transparent to that code.  On the other hand, maybe a respawn event might
+    * be useful, or maybe resend the currently non existant add event.  For
+    * consistancy with ecore_con, an add event is good anyway.
+    *
+    * The Ecore_exe structure is reused for respawning, so that the (opaque)
+    * pointer held by the user remains valid.  This means that the Ecore_Exe
+    * init and del functions may need to be split into two parts each to avoid
+    * duplicating code - common code part, and the rest.  This implies that
+    * the unchanging members mentioned next should NEVER change.
+    *
+    * These structure members don't need to change -
+    *   __list_data       - we stay on the list
+    *   ECORE_MAGIC       - this is a constant
+    *   data              - passed in originally
+    *   cmd               - passed in originally
+    *   flags             - passed in originally
+    *
+    * These structure members need to change -
+    *   tag               - state that must be regenerated, zap it
+    *   pid               - it will be different
+    *   child_fd_write    - it will be different
+    *   child_fd_read     - it will be different
+    *   child_fd_error    - it will be different
+    *   write_fd_handler  - we cannot change the fd used by a handler, this changes coz the fd changes.
+    *   read_fd_handler   - we cannot change the fd used by a handler, this changes coz the fd changes.
+    *   error_fd_handler  - we cannot change the fd used by a handler, this changes coz the fd changes.
+    *
+    * Hmm, the read, write, and error buffers could be tricky.
+    * They are not atomic, and could be in a semi complete state.
+    * They fall into the "state must be regenerated" mentioned above.
+    * A respawn/add event should take care of it.
+    *
+    * These structure members need to change -
+    *   write_data_buf    - state that must be regenerated, zap it
+    *   write_data_size   - state that must be regenerated, zap it
+    *   write_data_offset - state that must be regenerated, zap it
+    *   read_data_buf     - state that must be regenerated, zap it
+    *   read_data_size    - state that must be regenerated, zap it
+    *   error_data_buf    - state that must be regenerated, zap it
+    *   error_data_size   - state that must be regenerated, zap it
+    *   close_write       - state that must be regenerated, zap it
+    *
+    * There is the problem that an exe that fell over and needs respawning
+    * might keep falling over, keep needing to be respawned, and tie up system
+    * resources with the constant respawning.  An exponentially increasing
+    * timeout (with maximum timeout) between respawns should take care of that.
+    * Although this is not a "contention for a resource" problem, the exe falling
+    * over may be, so a random element added to the timeout may help, and won't
+    * hurt.  The user code may need to be informed that a timeout is in progress.
+    */
+
+struct _Ecore_Exe
+{
+   EINA_INLIST;
+   ECORE_MAGIC;
+   pid_t pid;
+   void *data;
+   char *tag, *cmd;
+   Ecore_Exe_Flags  flags;
+   Ecore_Fd_Handler *write_fd_handler; /* the fd_handler to handle write to child - if this was used, or NULL if not */
+   Ecore_Fd_Handler *read_fd_handler; /* the fd_handler to handle read from child - if this was used, or NULL if not */
+   Ecore_Fd_Handler *error_fd_handler; /* the fd_handler to handle errors from child - if this was used, or NULL if not */
+   void  *write_data_buf; /* a data buffer for data to write to the child -
+                                 * realloced as needed for more data and flushed when the fd handler says writes are possible
+				 */
+   int write_data_size; /* the size in bytes of the data buffer */
+   int write_data_offset; /* the offset in bytes in the data buffer */
+   void *read_data_buf; /* data read from the child awating delivery to an event */
+   int read_data_size; /* data read from child in bytes */
+   void *error_data_buf; /* errors read from the child awating delivery to an event */
+   int error_data_size; /* errors read from child in bytes */
+   int child_fd_write;	/* fd to write TO to send data to the child */
+   int child_fd_read;	/* fd to read FROM when child has sent us (the parent) data */
+   int child_fd_error;	/* fd to read FROM when child has sent us (the parent) errors */
+   int child_fd_write_x;	/* fd to write TO to send data to the child */
+   int child_fd_read_x;	/* fd to read FROM when child has sent us (the parent) data */
+   int child_fd_error_x;	/* fd to read FROM when child has sent us (the parent) errors */
+   Eina_Bool close_stdin : 1;
+
+   int start_bytes, end_bytes, start_lines, end_lines; /* Number of bytes/lines to auto pipe at start/end of stdout/stderr. */
+
+   Ecore_Timer *doomsday_clock; /* The Timer of Death.  Muahahahaha. */
+   void *doomsday_clock_dead; /* data for the doomsday clock */
+
+   void (*pre_free_cb)(void *data, const Ecore_Exe *exe);
+};
 
-#ifndef _WIN32
 
 /* TODO: Something to let people build a command line and does auto escaping -
  *
@@ -42,46 +139,36 @@
 
 struct _ecore_exe_dead_exe
 {
-   pid_t               pid;
-   char               *cmd;
+   pid_t pid;
+   char *cmd;
 };
 
-static inline void  _ecore_exe_exec_it(const char *exe_cmd, Ecore_Exe_Flags flags);
-
-static int          _ecore_exe_data_generic_handler(void *data,
-						    Ecore_Fd_Handler *
-						    fd_handler,
-						    Ecore_Exe_Flags flags);
-static int          _ecore_exe_data_error_handler(void *data,
-						  Ecore_Fd_Handler *
-						  fd_handler);
-static int          _ecore_exe_data_read_handler(void *data,
-						 Ecore_Fd_Handler * fd_handler);
-static int          _ecore_exe_data_write_handler(void *data,
-						  Ecore_Fd_Handler *
-						  fd_handler);
-static void         _ecore_exe_flush(Ecore_Exe * exe);
-static void         _ecore_exe_event_exe_data_free(void *data __UNUSED__,
-						   void *ev);
-static Ecore_Exe   *_ecore_exe_is_it_alive(pid_t pid);
-static int          _ecore_exe_make_sure_its_dead(void *data);
-static int          _ecore_exe_make_sure_its_really_dead(void *data);
+static inline void _ecore_exe_exec_it(const char *exe_cmd, Ecore_Exe_Flags flags);
+static Eina_Bool _ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler *fd_handler, Ecore_Exe_Flags flags);
+static Eina_Bool _ecore_exe_data_error_handler(void *data, Ecore_Fd_Handler *fd_handler);
+static Eina_Bool _ecore_exe_data_read_handler(void *data, Ecore_Fd_Handler *fd_handler);
+static Eina_Bool _ecore_exe_data_write_handler(void *data, Ecore_Fd_Handler *fd_handler);
+static void _ecore_exe_flush(Ecore_Exe * exe);
+static void _ecore_exe_event_exe_data_free(void *data __UNUSED__, void *ev);
+static Ecore_Exe *_ecore_exe_is_it_alive(pid_t pid);
+static Eina_Bool _ecore_exe_make_sure_its_dead(void *data);
+static Eina_Bool _ecore_exe_make_sure_its_really_dead(void *data);
 static Ecore_Exe_Event_Add *_ecore_exe_event_add_new(void);
-static void         _ecore_exe_event_add_free(void *data, void *ev);
-static void         _ecore_exe_dead_attach(Ecore_Exe *exe);
+static void _ecore_exe_event_add_free(void *data, void *ev);
+static void _ecore_exe_dead_attach(Ecore_Exe *exe);
 
-EAPI int            ECORE_EXE_EVENT_ADD = 0;
-EAPI int            ECORE_EXE_EVENT_DEL = 0;
-EAPI int            ECORE_EXE_EVENT_DATA = 0;
-EAPI int            ECORE_EXE_EVENT_ERROR = 0;
+EAPI int ECORE_EXE_EVENT_ADD = 0;
+EAPI int ECORE_EXE_EVENT_DEL = 0;
+EAPI int ECORE_EXE_EVENT_DATA = 0;
+EAPI int ECORE_EXE_EVENT_ERROR = 0;
 
-static Ecore_Exe   *exes = NULL;
-static const char  *shell = NULL;
+static Ecore_Exe *exes = NULL;
+static const char *shell = NULL;
 
 /* FIXME: This errno checking stuff should be put elsewhere for everybody to use.
  * For now it lives here though, just to make testing easier.
  */
-static int          _ecore_exe_check_errno(int result, const char *file, int line);
+static int _ecore_exe_check_errno(int result, const char *file, int line);
 
 #define E_IF_NO_ERRNO(result, foo, ok) \
   while (((ok) = _ecore_exe_check_errno( (result) = (foo), __FILE__, __LINE__)) == -1)   sleep(1); \
@@ -96,7 +183,7 @@ static int          _ecore_exe_check_errno(int result, const char *file, int lin
 static int
 _ecore_exe_check_errno(int result, const char *file, int line)
 {
-   int                 saved_errno = errno;
+   int saved_errno = errno;
 
    if (result == -1)
      {
@@ -158,65 +245,59 @@ _ecore_exe_check_errno(int result, const char *file, int line)
  */
 	switch (saved_errno)
 	  {
-	  case EACCES:
-	  case EAGAIN:
-	  case EINTR:
-	     {			/* Not now, try later. */
-		fprintf(stderr, "*** Must try again in %s @%u.\n", file, line);
-		result = -1;
-		break;
-	     }
-
-	  case EMFILE:
-	  case ENFILE:
-	  case ENOLCK:
-	     {			/* Low on resources. */
-		fprintf(stderr, "*** Low on resources in %s @%u.\n", file,
-			line);
-		result = 0;
-		break;
-	     }
-
-	  case EIO:
-	     {			/* I/O error. */
-		fprintf(stderr, "*** I/O error in %s @%u.\n", file, line);
-		result = 0;
-		break;
-	     }
-
-	  case EFAULT:
-	  case EBADF:
-	  case EINVAL:
-	  case EROFS:
-	  case EISDIR:
-	  case EDEADLK:
-	  case EPERM:
-	  case EBUSY:
-	     {			/* Programmer fucked up. */
-		fprintf(stderr,
-			"*** NAUGHTY PROGRAMMER!!!\n"
-			"*** SPANK SPANK SPANK!!!\n"
-			"*** Now go fix your code in %s @%u. Tut tut tut!\n"
-			"\n", file, line);
-		result = 0;
-		break;
-	     }
-
-	  default:
-	     {			/* Unsupported errno code, please add this one. */
-		fprintf(stderr,
-			"*** NAUGHTY PROGRAMMER!!!\n"
-			"*** SPANK SPANK SPANK!!!\n"
-			"*** Unsupported errno code %d, please add this one.\n"
-			"*** Now go fix your code in %s @%u, from %s @%u. Tut tut tut!\n"
-			"\n", saved_errno, __FILE__, __LINE__, file, line);
-		result = 0;
-		break;
-	     }
+           case EACCES:
+           case EAGAIN:
+           case EINTR:
+               { /* Not now, try later. */
+                  ERR("*** Must try again in %s @%u.", file, line);
+                  result = -1;
+                  break;
+               }
+           case EMFILE:
+           case ENFILE:
+           case ENOLCK:
+               { /* Low on resources. */
+                  ERR("*** Low on resources in %s @%u.", file,
+                      line);
+                  result = 0;
+                  break;
+               }
+           case EIO:
+               { /* I/O error. */
+                  ERR("*** I/O error in %s @%u.", file, line);
+                  result = 0;
+                  break;
+               }
+           case EFAULT:
+           case EBADF:
+           case EINVAL:
+           case EROFS:
+           case EISDIR:
+           case EDEADLK:
+           case EPERM:
+           case EBUSY:
+               { /* Programmer fucked up. */
+                  ERR("*** NAUGHTY PROGRAMMER!!!\n"
+                      "*** SPANK SPANK SPANK!!!\n"
+                      "*** Now go fix your code in %s @%u. Tut tut tut!",
+                      file, line);
+                  result = 0;
+                  break;
+               }
+           default:
+               { /* Unsupported errno code, please add this one. */
+                  ERR("*** NAUGHTY PROGRAMMER!!!\n"
+                      "*** SPANK SPANK SPANK!!!\n"
+                      "*** Unsupported errno code %d, please add this one.\n"
+                      "*** Now go fix your code in %s @%u, from %s @%u. Tut tut tut!",
+                      saved_errno, __FILE__, __LINE__, file, line);
+                  result = 0;
+                  break;
+               }
 	  }
      }
-   else				/* Everything is fine. */
-      result = 1;
+   else /* Everything is fine. */
+     result = 1;
 
    errno = saved_errno;
    return result;
@@ -234,10 +315,17 @@ static int run_pri = ECORE_EXE_PRIORITY_INHERIT;
  * Sets the priority at which to launch processes
  *
  * This sets the priority of processes run by ecore_exe_run() and
- * ecore_exe_pipe_run(). If set to ECORE_EXE_PRIORITY_INHERIT child processes
- * inherit the priority of their parent. This is the default.
+ * ecore_exe_pipe_run().
+ * @li On Windows, the child process is created by default with the
+ * #ECORE_EXE_WIN32_PRIORITY_NORMAL priority, unless the calling
+ * process is in #ECORE_EXE_WIN32_PRIORITY_IDLE or
+ * #ECORE_EXE_WIN32_PRIORITY_BELOW_NORMAL priority. In that case, the
+ * child process inherits this priority.
+ * @li On other platforms, if set to #ECORE_EXE_PRIORITY_INHERIT child
+ * processes inherits the priority of their parent. This is the default.
  *
- * @param   pri value -20 to 19 or ECORE_EXE_PRIORITY_INHERIT
+ * @param   pri value a Ecore_Exe_Win32_Priority value on Windows, -20
+ * to 19 or ECORE_EXE_PRIORITY_INHERIT on other OS.
  * @ingroup Ecore_Exe_Basic_Group
  */
 EAPI void
@@ -272,13 +360,13 @@ ecore_exe_run_priority_get(void)
  * @return  A process handle to the spawned process.
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI Ecore_Exe     *
+EAPI Ecore_Exe *
 ecore_exe_run(const char *exe_cmd, const void *data)
 {
 /* I'm just being paranoid again, leaving in the original code in case there is a problem. */
 #if 0
-   Ecore_Exe          *exe;
-   pid_t               pid;
+   Ecore_Exe *exe;
+   pid_t pid;
 
    if (!exe_cmd)
       return NULL;
@@ -332,17 +420,17 @@ ecore_exe_run(const char *exe_cmd, const void *data)
  * @return  A process handle to the spawned process.
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI Ecore_Exe     *
+EAPI Ecore_Exe *
 ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 {
-   Ecore_Exe          *exe = NULL;
-   int                 statusPipe[2] = { -1, -1 };
-   int                 errorPipe[2] = { -1, -1 };
-   int                 readPipe[2] = { -1, -1 };
-   int                 writePipe[2] = { -1, -1 };
-   int                 n = 0;
-   int                 ok = 1;
-   int                 result;
+   Ecore_Exe *exe = NULL;
+   int statusPipe[2] = { -1, -1 };
+   int errorPipe[2] = { -1, -1 };
+   int readPipe[2] = { -1, -1 };
+   int writePipe[2] = { -1, -1 };
+   int n = 0;
+   int ok = 1;
+   int result;
 
    if (!exe_cmd) return NULL;
    exe = calloc(1, sizeof(Ecore_Exe));
@@ -393,8 +481,8 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
      }
    if (ok)
      {
-	pid_t               pid = 0;
-	volatile int        vfork_exec_errno = 0;
+	pid_t pid = 0;
+	volatile int vfork_exec_errno = 0;
 
 	/* FIXME: I should double check this.  After a quick look around, this is already done, but via a more modern method. */
 	/* signal(SIGPIPE, SIG_IGN);    We only want EPIPE on errors */
@@ -402,7 +490,7 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 
 	if (pid == -1)
 	  {
-	     fprintf(stderr, "Failed to fork process\n");
+	     ERR("Failed to fork process");
 	     pid = 0;
 	  }
 	else if (pid == 0)	/* child */
@@ -410,9 +498,7 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 	     if (run_pri != ECORE_EXE_PRIORITY_INHERIT)
 	       {
 		  if ((run_pri >= -20) && (run_pri <= 19))
-		    {
-		       setpriority(PRIO_PROCESS, 0, run_pri);
-		    }
+                    setpriority(PRIO_PROCESS, 0, run_pri);
 	       }
 	     /* dup2 STDERR, STDIN, and STDOUT.  dup2() allegedly closes the
 	      * second pipe if it's open. On the other hand, there was the
@@ -458,7 +544,7 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 
 	     _exit(-1);
 	  }
-	else			/* parent */
+	else /* parent */
 	  {
 	     /* Close the unused pipes. */
 	     E_NO_ERRNO(result, close(statusPipe[1]), ok);
@@ -483,7 +569,7 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 	      * wrong */
 	     for (;;)
 	       {
-		  char                buf;
+		  char buf;
 
 		  E_NO_ERRNO(result, read(statusPipe[0], &buf, 1), ok);
 		  if (result == 0)
@@ -491,8 +577,7 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 		       if (vfork_exec_errno != 0)
 			 {
 			    n = vfork_exec_errno;
-			    fprintf(stderr, "Could not start \"%s\"\n",
-				    exe_cmd);
+			    ERR("Could not start \"%s\"", exe_cmd);
 			    pid = 0;
 			 }
 		       break;
@@ -582,7 +667,7 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 		       }
 		    }
 
-		  exes = _ecore_list2_append(exes, exe);
+		  exes = (Ecore_Exe *) eina_inlist_append(EINA_INLIST_GET(exes), EINA_INLIST_GET(exe));
 		  n = 0;
 	       }
 	     else
@@ -606,13 +691,38 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
 	if (e) /* Send the event. */
 	  ecore_event_add(ECORE_EXE_EVENT_ADD, e,
 			  _ecore_exe_event_add_free, NULL);
-	/* printf("Running as %d for %s.\n", exe->pid, exe->cmd); */
+	/* INF("Running as %d for %s.\n", exe->pid, exe->cmd); */
      }
 
    errno = n;
    return exe;
 }
 
+/**
+ * Defines a function to be called before really freeing the handle data.
+ *
+ * This might be useful for language bindings such as Python and Perl
+ * that need to deallocate wrappers associated with this handle.
+ *
+ * This handle should never be modified by this call. It should be
+ * considered informative only. All getters are valid when the given
+ * function is called back.
+ *
+ * @param exe The child process to attach the pre_free function.
+ * @param func The function to call before @a exe is freed.
+ */
+EAPI void
+ecore_exe_callback_pre_free_set(Ecore_Exe *exe, void (*func)(void *data, const Ecore_Exe *exe))
+{
+   if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
+     {
+	ECORE_MAGIC_FAIL(exe, ECORE_MAGIC_EXE,
+			 "ecore_exe_callback_pre_free_set");
+	return;
+     }
+   exe->pre_free_cb = func;
+}
+
 /**
  * Sends data to the given child process which it recieves on stdin.
  *
@@ -623,16 +733,36 @@ ecore_exe_pipe_run(const char *exe_cmd, Ecore_Exe_Flags flags, const void *data)
  * @param exe  The child process to send to
  * @param data The data to send
  * @param size The size of the data to send, in bytes
- * @return 1 if successful, 0 on failure.
+ * @return EINA_TRUE if successful, EINA_FALSE on failure.
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI int
-ecore_exe_send(Ecore_Exe * exe, void *data, int size)
+EAPI Eina_Bool
+ecore_exe_send(Ecore_Exe * exe, const void *data, int size)
 {
-   void               *buf;
+   void *buf;
+
+   if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
+     {
+	ECORE_MAGIC_FAIL(exe, ECORE_MAGIC_EXE, "ecore_exe_send");
+	return EINA_FALSE;
+     }
+
+   if (exe->close_stdin)
+     {
+	ERR("Ecore_Exe %p stdin is closed! Cannot send %d bytes from %p",
+	    exe, size, data);
+	return EINA_FALSE;
+     }
+
+   if (exe->child_fd_write == -1)
+     {
+	ERR("Ecore_Exe %p created without ECORE_EXE_PIPE_WRITE! "
+	    "Cannot send %d bytes from %p", exe, size, data);
+	return EINA_FALSE;
+     }
 
    buf = realloc(exe->write_data_buf, exe->write_data_size + size);
-   if (buf == NULL) return 0;
+   if (buf == NULL) return EINA_FALSE;
 
    exe->write_data_buf = buf;
    memcpy((char *)exe->write_data_buf + exe->write_data_size, data, size);
@@ -641,7 +771,7 @@ ecore_exe_send(Ecore_Exe * exe, void *data, int size)
    if (exe->write_fd_handler)
       ecore_main_fd_handler_active_set(exe->write_fd_handler, ECORE_FD_WRITE);
 
-   return 1;
+   return EINA_TRUE;
 }
 
 /**
@@ -651,7 +781,7 @@ ecore_exe_send(Ecore_Exe * exe, void *data, int size)
  * @ingroup Ecore_Exe_Basic_Group
  */
 EAPI void
-ecore_exe_close_stdin(Ecore_Exe * exe)
+ecore_exe_close_stdin(Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -662,7 +792,8 @@ ecore_exe_close_stdin(Ecore_Exe * exe)
 }
 
 /**
- * Sets the auto pipe limits for the given process handle
+ * Sets the auto pipe limits for the given process handle. On Windows
+ * this function does nothing.
  *
  * @param   exe The given process handle.
  * @param   start_bytes limit of bytes at start of output to buffer.
@@ -672,8 +803,7 @@ ecore_exe_close_stdin(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Basic_Group
  */
 EAPI void
-ecore_exe_auto_limits_set(Ecore_Exe * exe, int start_bytes, int end_bytes,
-			  int start_lines, int end_lines)
+ecore_exe_auto_limits_set(Ecore_Exe *exe, int start_bytes, int end_bytes, int start_lines, int end_lines)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -734,12 +864,12 @@ ecore_exe_auto_limits_set(Ecore_Exe * exe, int start_bytes, int end_bytes,
  * @ingroup Ecore_Exe_Basic_Group
  */
 EAPI Ecore_Exe_Event_Data *
-ecore_exe_event_data_get(Ecore_Exe * exe, Ecore_Exe_Flags flags)
+ecore_exe_event_data_get(Ecore_Exe *exe, Ecore_Exe_Flags flags)
 {
    Ecore_Exe_Event_Data *e = NULL;
-   int                 is_buffered = 0;
-   unsigned char      *inbuf;
-   int                 inbuf_num;
+   int is_buffered = 0;
+   unsigned char *inbuf;
+   int inbuf_num;
 
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -786,11 +916,11 @@ ecore_exe_event_data_get(Ecore_Exe * exe, Ecore_Exe_Flags flags)
 
 	if (is_buffered)
 	  {			/* Deal with line buffering. */
-	     int                 max = 0;
-	     int                 count = 0;
-	     int                 i;
-	     int                 last = 0;
-	     char               *c;
+	     int max = 0;
+	     int count = 0;
+	     int i;
+	     int last = 0;
+	     char *c;
 
 	     c = (char *)inbuf;
 	     for (i = 0; i < inbuf_num; i++)	/* Find the lines. */
@@ -837,8 +967,7 @@ ecore_exe_event_data_get(Ecore_Exe * exe, Ecore_Exe_Flags flags)
 	       }
 	     if (i > last)	/* Partial line left over, save it for next time. */
 	       {
-	          if (e)
-		     e->size = last;
+	          if (e) e->size = last;
 		  if (flags & ECORE_EXE_PIPE_READ)
 		    {
 		       exe->read_data_size = i - last;
@@ -866,7 +995,7 @@ ecore_exe_event_data_get(Ecore_Exe * exe, Ecore_Exe_Flags flags)
  * @ingroup Ecore_Exe_Basic_Group
  */
 EAPI void
-ecore_exe_tag_set(Ecore_Exe * exe, const char *tag)
+ecore_exe_tag_set(Ecore_Exe *exe, const char *tag)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -876,6 +1005,8 @@ ecore_exe_tag_set(Ecore_Exe * exe, const char *tag)
    IF_FREE(exe->tag);
    if (tag)
       exe->tag = strdup(tag);
+   else
+      exe->tag = NULL;
 }
 
 /**
@@ -885,11 +1016,13 @@ ecore_exe_tag_set(Ecore_Exe * exe, const char *tag)
  * else on this @p exe.
  *
  * @param   exe The given process handle.
- * @return  The string attached to @p exe.
+ * @return The string attached to @p exe. It is a handle to existing
+ *         internal string and should not be modified, use
+ *         ecore_exe_tag_set() to change it. It might be @c NULL.
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI char          *
-ecore_exe_tag_get(Ecore_Exe * exe)
+EAPI const char *
+ecore_exe_tag_get(const Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -910,12 +1043,12 @@ ecore_exe_tag_get(Ecore_Exe * exe)
  *          called.
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI void          *
-ecore_exe_free(Ecore_Exe * exe)
+EAPI void *
+ecore_exe_free(Ecore_Exe *exe)
 {
-   void               *data;
-   int                 ok = 0;
-   int                 result;
+   void *data;
+   int ok = 0;
+   int result;
 
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -925,6 +1058,9 @@ ecore_exe_free(Ecore_Exe * exe)
 
    data = exe->data;
 
+   if (exe->pre_free_cb)
+     exe->pre_free_cb(data, exe);
+
    if (exe->doomsday_clock)
      {
 	struct _ecore_exe_dead_exe *dead;
@@ -959,7 +1095,7 @@ ecore_exe_free(Ecore_Exe * exe)
    IF_FREE(exe->error_data_buf);
    IF_FREE(exe->cmd);
 
-   exes = _ecore_list2_remove(exes, exe);
+   exes = (Ecore_Exe *) eina_inlist_remove(EINA_INLIST_GET(exes), EINA_INLIST_GET(exe));
    ECORE_MAGIC_SET(exe, ECORE_MAGIC_NONE);
    IF_FREE(exe->tag);
    free(exe);
@@ -973,8 +1109,9 @@ ecore_exe_free(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Basic_Group
  */
 EAPI void
-ecore_exe_event_data_free(Ecore_Exe_Event_Data * e)
+ecore_exe_event_data_free(Ecore_Exe_Event_Data *e)
 {
+   if (!e) return;
    IF_FREE(e->lines);
    IF_FREE(e->data);
    free(e);
@@ -986,8 +1123,8 @@ ecore_exe_event_data_free(Ecore_Exe_Event_Data * e)
  * @return  The process ID on success.  @c -1 otherwise.
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI                pid_t
-ecore_exe_pid_get(Ecore_Exe * exe)
+EAPI pid_t
+ecore_exe_pid_get(const Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1000,11 +1137,13 @@ ecore_exe_pid_get(Ecore_Exe * exe)
 /**
  * Retrieves the command of the given spawned process.
  * @param   exe Handle to the given spawned process.
- * @return  The command on success.  NULL otherwise.
+ * @return The command on success.  NULL otherwise. This string is the
+ *         pointer to the internal value and must not be modified in
+ *         any way.
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI char          *
-ecore_exe_cmd_get(Ecore_Exe * exe)
+EAPI const char *
+ecore_exe_cmd_get(const Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1017,11 +1156,12 @@ ecore_exe_cmd_get(Ecore_Exe * exe)
 /**
  * Retrieves the data attached to the given process handle.
  * @param   exe The given process handle.
- * @return  The data pointer attached to @p exe.
+ * @return The data pointer attached to @p exe Given to
+ *         ecore_exe_run() or ecore_exe_pipe_run()
  * @ingroup Ecore_Exe_Basic_Group
  */
-EAPI void          *
-ecore_exe_data_get(Ecore_Exe * exe)
+EAPI void *
+ecore_exe_data_get(const Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1031,6 +1171,23 @@ ecore_exe_data_get(Ecore_Exe * exe)
    return exe->data;
 }
 
+/**
+ * Retrieves the flags attached to the given process handle.
+ * @param   exe The given process handle.
+ * @return  The flags attached to @p exe.
+ * @ingroup Ecore_Exe_Basic_Group
+ */
+EAPI Ecore_Exe_Flags
+ecore_exe_flags_get(const Ecore_Exe *exe)
+{
+   if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
+     {
+	ECORE_MAGIC_FAIL(exe, ECORE_MAGIC_EXE, "ecore_exe_data_get");
+	return 0;
+     }
+   return exe->flags;
+}
+
 /**
  * @defgroup Ecore_Exe_Signal_Group Spawned Process Signal Functions
  *
@@ -1043,7 +1200,7 @@ ecore_exe_data_get(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_pause(Ecore_Exe * exe)
+ecore_exe_pause(Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1059,7 +1216,7 @@ ecore_exe_pause(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_continue(Ecore_Exe * exe)
+ecore_exe_continue(Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1075,7 +1232,7 @@ ecore_exe_continue(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_interrupt(Ecore_Exe * exe)
+ecore_exe_interrupt(Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1092,7 +1249,7 @@ ecore_exe_interrupt(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_quit(Ecore_Exe * exe)
+ecore_exe_quit(Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1109,7 +1266,7 @@ ecore_exe_quit(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_terminate(Ecore_Exe * exe)
+ecore_exe_terminate(Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1117,7 +1274,7 @@ ecore_exe_terminate(Ecore_Exe * exe)
 	return;
      }
    _ecore_exe_dead_attach(exe);
-   printf("Sending TERM signal to %s (%d).\n", exe->cmd, exe->pid);
+   INF("Sending TERM signal to %s (%d).", exe->cmd, exe->pid);
    kill(exe->pid, SIGTERM);
 }
 
@@ -1127,7 +1284,7 @@ ecore_exe_terminate(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_kill(Ecore_Exe * exe)
+ecore_exe_kill(Ecore_Exe *exe)
 {
    struct _ecore_exe_dead_exe *dead;
 
@@ -1147,7 +1304,7 @@ ecore_exe_kill(Ecore_Exe * exe)
 	   ecore_timer_add(10.0, _ecore_exe_make_sure_its_really_dead, dead);
      }
 
-   printf("Sending KILL signal to %s (%d).\n", exe->cmd, exe->pid);
+   INF("Sending KILL signal to %s (%d).", exe->cmd, exe->pid);
    kill(exe->pid, SIGKILL);
 }
 
@@ -1159,7 +1316,7 @@ ecore_exe_kill(Ecore_Exe * exe)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_signal(Ecore_Exe * exe, int num)
+ecore_exe_signal(Ecore_Exe *exe, int num)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1178,7 +1335,7 @@ ecore_exe_signal(Ecore_Exe * exe, int num)
  * @ingroup Ecore_Exe_Signal_Group
  */
 EAPI void
-ecore_exe_hup(Ecore_Exe * exe)
+ecore_exe_hup(Ecore_Exe *exe)
 {
    if (!ECORE_MAGIC_CHECK(exe, ECORE_MAGIC_EXE))
      {
@@ -1188,10 +1345,10 @@ ecore_exe_hup(Ecore_Exe * exe)
    kill(exe->pid, SIGHUP);
 }
 
-static Ecore_Exe   *
+static Ecore_Exe *
 _ecore_exe_is_it_alive(pid_t pid)
 {
-   Ecore_Exe          *exe = NULL;
+   Ecore_Exe *exe = NULL;
 
    /* FIXME: There is no nice, safe, OS independant way to tell if a
     * particular PID is still alive.  I have written code to do so
@@ -1225,7 +1382,7 @@ _ecore_exe_is_it_alive(pid_t pid)
    return exe;
 }
 
-static int
+static Eina_Bool
 _ecore_exe_make_sure_its_dead(void *data)
 {
    struct _ecore_exe_dead_exe *dead;
@@ -1233,16 +1390,16 @@ _ecore_exe_make_sure_its_dead(void *data)
    dead = data;
    if (dead)
      {
-	Ecore_Exe          *exe = NULL;
+	Ecore_Exe *exe = NULL;
 
 	if ((exe = _ecore_exe_is_it_alive(dead->pid)) != NULL)
 	  {
 	     if (dead->cmd)
-		printf("Sending KILL signal to alledgedly dead %s (%d).\n",
+		INF("Sending KILL signal to alledgedly dead %s (%d).",
 		       dead->cmd, dead->pid);
 	     else
-		printf("Sending KILL signal to alledgedly dead PID %d.\n",
-		       dead->pid);
+		INF("Sending KILL signal to alledgedly dead PID %d.",
+		    dead->pid);
 	     exe->doomsday_clock =
 		ecore_timer_add(10.0, _ecore_exe_make_sure_its_really_dead,
 				dead);
@@ -1254,10 +1411,10 @@ _ecore_exe_make_sure_its_dead(void *data)
 	     free(dead);
 	  }
      }
-   return 0;
+   return ECORE_CALLBACK_CANCEL;
 }
 
-static int
+static Eina_Bool
 _ecore_exe_make_sure_its_really_dead(void *data)
 {
    struct _ecore_exe_dead_exe *dead;
@@ -1265,22 +1422,21 @@ _ecore_exe_make_sure_its_really_dead(void *data)
    dead = data;
    if (dead)
      {
-	Ecore_Exe          *exe = NULL;
+	Ecore_Exe *exe = NULL;
 
 	if ((exe = _ecore_exe_is_it_alive(dead->pid)) != NULL)
 	  {
-	     printf
-		("RUN!  The zombie wants to eat your brains!  And your CPU!\n");
+	     ERR("RUN!  The zombie wants to eat your brains!  And your CPU!");
 	     if (dead->cmd)
-		printf("%s (%d) is not really dead.\n", dead->cmd, dead->pid);
+		INF("%s (%d) is not really dead.", dead->cmd, dead->pid);
 	     else
-		printf("PID %d is not really dead.\n", dead->pid);
+		INF("PID %d is not really dead.", dead->pid);
 	     exe->doomsday_clock = NULL;
 	  }
 	IF_FREE(dead->cmd);
 	free(dead);
      }
-   return 0;
+   return ECORE_CALLBACK_CANCEL;
 }
 
 void
@@ -1296,32 +1452,41 @@ void
 _ecore_exe_shutdown(void)
 {
    while (exes)
-      ecore_exe_free(exes);
+     ecore_exe_free(exes);
 }
 
-Ecore_Exe          *
+Ecore_Exe *
 _ecore_exe_find(pid_t pid)
 {
-   Ecore_List2        *l;
+   Ecore_Exe *exe;
 
-   for (l = (Ecore_List2 *) exes; l; l = l->next)
+   EINA_INLIST_FOREACH(exes, exe)
      {
-	Ecore_Exe          *exe;
-
-	exe = (Ecore_Exe *) l;
 	if (exe->pid == pid)
 	   return exe;
      }
    return NULL;
 }
 
+Ecore_Timer *
+_ecore_exe_doomsday_clock_get(Ecore_Exe *exe)
+{
+   return exe->doomsday_clock;
+}
+
+void
+_ecore_exe_doomsday_clock_set(Ecore_Exe *exe, Ecore_Timer *dc)
+{
+   exe->doomsday_clock = dc;
+}
+
 static inline void
 _ecore_exe_exec_it(const char *exe_cmd, Ecore_Exe_Flags flags)
 {
-   char                use_sh = 1;
-   char               *buf = NULL;
-   char              **args = NULL;
-   int                 save_errno = 0;
+   char use_sh = 1;
+   char *buf = NULL;
+   char **args = NULL;
+   int  save_errno = 0;
 
    /* So what is this doing?
     *
@@ -1331,9 +1496,9 @@ _ecore_exe_exec_it(const char *exe_cmd, Ecore_Exe_Flags flags)
     */
    if (!strpbrk(exe_cmd, "|&;<>()$`\\\"'*?#"))
      {
-	char               *token;
-	char                pre_command = 1;
-	int                 num_tokens = 0;
+	char *token;
+	char pre_command = 1;
+	int num_tokens = 0;
 
 	if (!(buf = strdup(exe_cmd)))
 	   return;
@@ -1358,7 +1523,7 @@ _ecore_exe_exec_it(const char *exe_cmd, Ecore_Exe_Flags flags)
 	IF_FREE(buf);
 	if ((!token) && (num_tokens))
 	  {
-	     int                 i = 0;
+	     int i = 0;
 
 	     if (!(buf = strdup(exe_cmd)))
 	       return;
@@ -1410,14 +1575,12 @@ _ecore_exe_exec_it(const char *exe_cmd, Ecore_Exe_Flags flags)
    return;
 }
 
-static int
-_ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler * fd_handler,
-				Ecore_Exe_Flags flags)
+static Eina_Bool
+_ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler *fd_handler, Ecore_Exe_Flags flags)
 {
-   Ecore_Exe          *exe;
-   int                 child_fd;
-   int                 is_buffered = 0;
-   int                 event_type;
+   Ecore_Exe *exe;
+   int child_fd;
+   int event_type;
 
    exe = data;
 
@@ -1427,23 +1590,19 @@ _ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler * fd_handler,
 	flags = ECORE_EXE_PIPE_READ;
 	event_type = ECORE_EXE_EVENT_DATA;
 	child_fd = exe->child_fd_read;
-	if (exe->flags & ECORE_EXE_PIPE_READ_LINE_BUFFERED)
-	   is_buffered = 1;
      }
    else
      {
 	flags = ECORE_EXE_PIPE_ERROR;
 	event_type = ECORE_EXE_EVENT_ERROR;
 	child_fd = exe->child_fd_error;
-	if (exe->flags & ECORE_EXE_PIPE_ERROR_LINE_BUFFERED)
-	   is_buffered = 1;
      }
 
    if ((fd_handler)
        && (ecore_main_fd_handler_active_get(fd_handler, ECORE_FD_READ)))
      {
-	unsigned char      *inbuf;
-	int                 inbuf_num;
+	unsigned char *inbuf;
+	int inbuf_num;
 
 	/* Get any left over data from last time. */
 	if (flags & ECORE_EXE_PIPE_READ)
@@ -1463,8 +1622,8 @@ _ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler * fd_handler,
 
 	for (;;)
 	  {
-	     int                 num, lost_exe;
-	     char                buf[READBUFSIZ];
+	     int num, lost_exe;
+	     char buf[READBUFSIZ];
 
 	     lost_exe = 0;
 	     errno = 0;
@@ -1508,9 +1667,9 @@ _ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler * fd_handler,
 			 {
 			    e = ecore_exe_event_data_get(exe, flags);
 			    if (e)	/* Send the event. */
-			       ecore_event_add(event_type, e,
-					       _ecore_exe_event_exe_data_free,
-					       NULL);
+                              ecore_event_add(event_type, e,
+                                              _ecore_exe_event_exe_data_free,
+                                              NULL);
 			 }
 		    }
 		  if (lost_exe)
@@ -1518,16 +1677,14 @@ _ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler * fd_handler,
 		       if (flags & ECORE_EXE_PIPE_READ)
 			 {
 			    if (exe->read_data_size)
-			       printf
-				  ("There are %d bytes left unsent from the dead exe %s.\n",
-				   exe->read_data_size, exe->cmd);
+                              INF("There are %d bytes left unsent from the dead exe %s.",
+                                  exe->read_data_size, exe->cmd);
 			 }
 		       else
 			 {
 			    if (exe->error_data_size)
-			       printf
-				  ("There are %d bytes left unsent from the dead exe %s.\n",
-				   exe->error_data_size, exe->cmd);
+                              INF("There are %d bytes left unsent from the dead exe %s.",
+                                  exe->error_data_size, exe->cmd);
 			 }
 		       /* Thought about this a bit.  If the exe has actually
 		        * died, this won't do any harm as it must have died
@@ -1543,43 +1700,42 @@ _ecore_exe_data_generic_handler(void *data, Ecore_Fd_Handler * fd_handler,
 	  }
      }
 
-   return 1;
+   return ECORE_CALLBACK_RENEW;
 }
 
-static int
-_ecore_exe_data_error_handler(void *data, Ecore_Fd_Handler * fd_handler)
+static Eina_Bool
+_ecore_exe_data_error_handler(void *data, Ecore_Fd_Handler *fd_handler)
 {
    return _ecore_exe_data_generic_handler(data, fd_handler,
 					  ECORE_EXE_PIPE_ERROR);
 }
 
-static int
-_ecore_exe_data_read_handler(void *data, Ecore_Fd_Handler * fd_handler)
+static Eina_Bool
+_ecore_exe_data_read_handler(void *data, Ecore_Fd_Handler *fd_handler)
 {
    return _ecore_exe_data_generic_handler(data, fd_handler,
 					  ECORE_EXE_PIPE_READ);
 }
 
-static int
-_ecore_exe_data_write_handler(void *data, Ecore_Fd_Handler * fd_handler __UNUSED__)
+static Eina_Bool
+_ecore_exe_data_write_handler(void *data, Ecore_Fd_Handler *fd_handler __UNUSED__)
 {
-   Ecore_Exe          *exe;
+   Ecore_Exe *exe;
 
    exe = data;
-   if ((exe->write_fd_handler)
-       &&
+   if ((exe->write_fd_handler) &&
        (ecore_main_fd_handler_active_get
 	(exe->write_fd_handler, ECORE_FD_WRITE)))
-      _ecore_exe_flush(exe);
+     _ecore_exe_flush(exe);
 
    /* If we have sent all there is to send, and we need to close the pipe, then close it. */
    if ((exe->close_stdin == 1)
        && (exe->write_data_size == exe->write_data_offset))
      {
-	int                 ok = 0;
-	int                 result;
+	int ok = 0;
+	int result;
 
-	printf("Closing stdin for %s\n", exe->cmd);
+	INF("Closing stdin for %s", exe->cmd);
 	/* if (exe->child_fd_write != -1)  E_NO_ERRNO(result, fsync(exe->child_fd_write), ok);   This a) doesn't work, and b) isn't needed. */
 	IF_FN_DEL(ecore_main_fd_handler_del, exe->write_fd_handler);
 	if (exe->child_fd_write != -1)
@@ -1588,13 +1744,13 @@ _ecore_exe_data_write_handler(void *data, Ecore_Fd_Handler * fd_handler __UNUSED
 	IF_FREE(exe->write_data_buf);
      }
 
-   return 1;
+   return ECORE_CALLBACK_RENEW;
 }
 
 static void
-_ecore_exe_flush(Ecore_Exe * exe)
+_ecore_exe_flush(Ecore_Exe *exe)
 {
-   int                 count;
+   int count;
 
    /* check whether we need to write anything at all. */
    if ((exe->child_fd_write == -1) || (!exe->write_data_buf))
@@ -1671,7 +1827,7 @@ _ecore_exe_event_del_free(void *data __UNUSED__, void *ev)
 
    e = ev;
    if (e->exe)
-      ecore_exe_free(e->exe);
+     ecore_exe_free(e->exe);
    free(e);
 }
 
@@ -1692,5 +1848,3 @@ _ecore_exe_dead_attach(Ecore_Exe *exe)
 	exe->doomsday_clock_dead = dead;
      }
 }
-
-#endif
